@@ -17,16 +17,22 @@
 /* Anyone with a 10M config file is doing something very wrong */
 #define MAX_CONFIG_SIZE (10*1024*1024)
 
+struct uint32_percent {
+	uint32_t value;
+	bool percent;
+};
+
 static char *strip(const char *str);
 static bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const char *option, const char *value);
 static char *get_config_path(void);
+static uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent, uint32_t scale, bool use_scale);
 
 static uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool *err);
 static bool parse_bool(const char *filename, size_t lineno, const char *str, bool *err);
 static struct color parse_color(const char *filename, size_t lineno, const char *str, bool *err);
 static uint32_t parse_uint32(const char *filename, size_t lineno, const char *str, bool *err);
 static int32_t parse_int32(const char *filename, size_t lineno, const char *str, bool *err);
-static uint32_t parse_uint32_percent(const char *filename, size_t lineno, const char *str, bool *err, uint32_t max);
+static struct uint32_percent parse_uint32_percent(const char *filename, size_t lineno, const char *str, bool *err);
 
 /*
  * Function-like macro. Yuck.
@@ -231,6 +237,7 @@ char *strip(const char *str)
 bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const char *option, const char *value)
 {
 	bool err = false;
+	struct uint32_percent percent;
 	if (strcasecmp(option, "anchor") == 0) {
 		tofi->anchor = parse_anchor(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "background-color") == 0) {
@@ -263,38 +270,73 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 		tofi->window.entry.foreground_color = parse_color(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "selection-color") == 0) {
 		tofi->window.entry.selection_foreground_color = parse_color(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "selection-match-color") == 0) {
+		tofi->window.entry.selection_highlight_color = parse_color(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "selection-padding") == 0) {
+		tofi->window.entry.selection_background_padding = parse_int32(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "selection-background") == 0) {
 		tofi->window.entry.selection_background_color = parse_color(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "width") == 0) {
-		tofi->window.width = parse_uint32_percent(filename, lineno, value, &err, tofi->output_width);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.width = percent.value;
+		tofi->window.width_is_percent = percent.percent;
 	} else if (strcasecmp(option, "height") == 0) {
-		tofi->window.height = parse_uint32_percent(filename, lineno, value, &err, tofi->output_height);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.height = percent.value;
+		tofi->window.height_is_percent = percent.percent;
 	} else if (strcasecmp(option, "margin-top") == 0) {
-		tofi->window.margin_top = parse_uint32_percent(filename, lineno, value, &err, tofi->output_height);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.margin_top = percent.value;
+		tofi->window.margin_top_is_percent = percent.percent;
 	} else if (strcasecmp(option, "margin-bottom") == 0) {
-		tofi->window.margin_bottom = parse_uint32_percent(filename, lineno, value, &err, tofi->output_height);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.margin_bottom = percent.value;
+		tofi->window.margin_bottom_is_percent = percent.percent;
 	} else if (strcasecmp(option, "margin-left") == 0) {
-		tofi->window.margin_left = parse_uint32_percent(filename, lineno, value, &err, tofi->output_width);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.margin_left = percent.value;
+		tofi->window.margin_left_is_percent = percent.percent;
 	} else if (strcasecmp(option, "margin-right") == 0) {
-		tofi->window.margin_right = parse_uint32_percent(filename, lineno, value, &err, tofi->output_width);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.margin_right = percent.value;
+		tofi->window.margin_right_is_percent = percent.percent;
 	} else if (strcasecmp(option, "padding-top") == 0) {
-		tofi->window.entry.padding_top = parse_uint32_percent(filename, lineno, value, &err, tofi->output_height);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.entry.padding_top = percent.value;
+		tofi->window.entry.padding_top_is_percent = percent.percent;
 	} else if (strcasecmp(option, "padding-bottom") == 0) {
-		tofi->window.entry.padding_bottom = parse_uint32_percent(filename, lineno, value, &err, tofi->output_height);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.entry.padding_bottom = percent.value;
+		tofi->window.entry.padding_bottom_is_percent = percent.percent;
 	} else if (strcasecmp(option, "padding-left") == 0) {
-		tofi->window.entry.padding_left = parse_uint32_percent(filename, lineno, value, &err, tofi->output_width);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.entry.padding_left = percent.value;
+		tofi->window.entry.padding_left_is_percent = percent.percent;
 	} else if (strcasecmp(option, "padding-right") == 0) {
-		tofi->window.entry.padding_right = parse_uint32_percent(filename, lineno, value, &err, tofi->output_width);
+		percent = parse_uint32_percent(filename, lineno, value, &err);
+		tofi->window.entry.padding_right = percent.value;
+		tofi->window.entry.padding_right_is_percent = percent.percent;
 	} else if (strcasecmp(option, "horizontal") == 0) {
 		tofi->window.entry.horizontal = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "hide-cursor") == 0) {
 		tofi->hide_cursor = parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "history") == 0) {
 		tofi->use_history = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "fuzzy-match") == 0) {
+		tofi->fuzzy_match = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "drun-launch") == 0) {
+		tofi->drun_launch = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "drun-print-exec") == 0) {
+		log_warning("drun-print-exec is deprecated, as it is now always true.\n"
+				"           This option may be removed in a future version of tofi.\n");
 	} else if (strcasecmp(option, "hint-font") == 0) {
 		tofi->window.entry.harfbuzz.disable_hinting = !parse_bool(filename, lineno, value, &err);
 	} else if (strcasecmp(option, "late-keyboard-init") == 0) {
 		tofi->late_keyboard_init = parse_bool(filename, lineno, value, &err);
+	} else if (strcasecmp(option, "output") == 0) {
+		snprintf(tofi->target_output_name, N_ELEM(tofi->target_output_name), "%s", value);
+	} else if (strcasecmp(option, "scale") == 0) {
+		tofi->use_scale = parse_bool(filename, lineno, value, &err);
 	} else {
 		PARSE_ERROR(filename, lineno, "Unknown option \"%s\"\n", option);
 		err = true;
@@ -302,13 +344,109 @@ bool parse_option(struct tofi *tofi, const char *filename, size_t lineno, const
 	return !err;
 }
 
-void apply_option(struct tofi *tofi, const char *option, const char *value)
+void config_apply(struct tofi *tofi, const char *option, const char *value)
 {
 	if (!parse_option(tofi, "", 0, option, value)) {
 		exit(EXIT_FAILURE);
 	};
 }
 
+uint32_t fixup_percentage(uint32_t value, uint32_t base, bool is_percent, uint32_t scale, bool use_scale)
+{
+	if (is_percent) {
+		return value * base / 100;
+	} else if (use_scale) {
+		return value * scale;
+	}
+	return value;
+}
+
+void config_fixup_values(struct tofi *tofi)
+{
+	uint32_t scale = tofi->window.scale;
+
+	/*
+	 * Scale fonts to the correct size.
+	 *
+	 * TODO: In the next release (0.6.0), this should be moved within
+	 * the use_scale conditional.
+	 */
+	tofi->window.entry.font_size *= scale;
+
+	if (tofi->use_scale) {
+		struct entry *entry = &tofi->window.entry;
+
+		entry->corner_radius *= scale;
+		entry->selection_background_padding *= scale;
+		entry->result_spacing *= scale;
+		entry->input_width *= scale;
+		entry->outline_width *= scale;
+		entry->border_width *= scale;
+	}
+
+	/* These values should only be scaled if they're not percentages. */
+	tofi->window.width = fixup_percentage(
+			tofi->window.width,
+			tofi->output_width,
+			tofi->window.width_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.height = fixup_percentage(
+			tofi->window.height,
+			tofi->output_height,
+			tofi->window.height_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.margin_top = fixup_percentage(
+			tofi->window.margin_top,
+			tofi->output_height,
+			tofi->window.margin_top_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.margin_bottom = fixup_percentage(
+			tofi->window.margin_bottom,
+			tofi->output_height,
+			tofi->window.margin_bottom_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.margin_left = fixup_percentage(
+			tofi->window.margin_left,
+			tofi->output_width,
+			tofi->window.margin_left_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.margin_right = fixup_percentage(
+			tofi->window.margin_right,
+			tofi->output_width,
+			tofi->window.margin_right_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.entry.padding_top = fixup_percentage(
+			tofi->window.entry.padding_top,
+			tofi->output_height,
+			tofi->window.entry.padding_top_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.entry.padding_bottom = fixup_percentage(
+			tofi->window.entry.padding_bottom,
+			tofi->output_height,
+			tofi->window.entry.padding_bottom_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.entry.padding_left = fixup_percentage(
+			tofi->window.entry.padding_left,
+			tofi->output_width,
+			tofi->window.entry.padding_left_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+	tofi->window.entry.padding_right = fixup_percentage(
+			tofi->window.entry.padding_right,
+			tofi->output_width,
+			tofi->window.entry.padding_right_is_percent,
+			tofi->window.scale,
+			tofi->use_scale);
+}
+
 char *get_config_path()
 {
 	char *base_dir = getenv("XDG_CONFIG_HOME");
@@ -393,7 +531,14 @@ uint32_t parse_anchor(const char *filename, size_t lineno, const char *str, bool
 
 struct color parse_color(const char *filename, size_t lineno, const char *str, bool *err)
 {
-	return hex_to_color(str);
+	struct color color = hex_to_color(str);
+	if (color.r == -1) {
+		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as a color.\n", str);
+		if (err) {
+			*err = true;
+		}
+	}
+	return color;
 }
 
 uint32_t parse_uint32(const char *filename, size_t lineno, const char *str, bool *err)
@@ -434,17 +579,18 @@ int32_t parse_int32(const char *filename, size_t lineno, const char *str, bool *
 	return ret;
 }
 
-uint32_t parse_uint32_percent(const char *filename, size_t lineno, const char *str, bool *err, uint32_t max)
+struct uint32_percent parse_uint32_percent(const char *filename, size_t lineno, const char *str, bool *err)
 {
 	errno = 0;
 	char *endptr;
-	int32_t ret = strtoul(str, &endptr, 0);
+	int32_t val = strtoul(str, &endptr, 0);
+	bool percent = false;
 	if (endptr == str) {
 		PARSE_ERROR(filename, lineno, "Failed to parse \"%s\" as unsigned int.\n", str);
 		if (err) {
 			*err = true;
 		}
-	} else if (errno || ret < 0) {
+	} else if (errno || val < 0) {
 		PARSE_ERROR(filename, lineno, "Unsigned int value \"%s\" out of range.\n", str);
 		if (err) {
 			*err = true;
@@ -452,8 +598,8 @@ uint32_t parse_uint32_percent(const char *filename, size_t lineno, const char *s
 	}
 	if (!err || !*err) {
 		if (*endptr == '%') {
-			ret = max * ret / 100;
+			percent = true;
 		}
 	}
-	return ret;
+	return (struct uint32_percent){ val, percent };
 }
