@@ -7,6 +7,7 @@
 #include <cairo/cairo.h>
 #include <wchar.h>
 #include "color.h"
+#include "desktop_vec.h"
 #include "history.h"
 #include "image.h"
 #include "surface.h"
@@ -33,14 +34,24 @@ struct entry {
 	uint32_t input_mb_length;
 
 	uint32_t selection;
+	uint32_t first_result;
 	struct string_vec results;
 	struct string_vec commands;
+	struct desktop_vec apps;
 	struct history history;
 	bool use_pango;
 
+	uint32_t clip_x;
+	uint32_t clip_y;
+	uint32_t clip_width;
+	uint32_t clip_height;
+
 	/* Options */
+	bool drun;
 	bool horizontal;
 	uint32_t num_results;
+	uint32_t num_results_drawn;
+	uint32_t last_num_results_drawn;
 	int32_t result_spacing;
 	uint32_t font_size;
 	char font_name[MAX_FONT_NAME_LENGTH];
@@ -50,11 +61,17 @@ struct entry {
 	uint32_t padding_bottom;
 	uint32_t padding_left;
 	uint32_t padding_right;
+	bool padding_top_is_percent;
+	bool padding_bottom_is_percent;
+	bool padding_left_is_percent;
+	bool padding_right_is_percent;
+	int32_t selection_background_padding;
 	uint32_t input_width;
 	uint32_t border_width;
 	uint32_t outline_width;
 	struct color foreground_color;
 	struct color background_color;
+	struct color selection_highlight_color;
 	struct color selection_foreground_color;
 	struct color selection_background_color;
 	struct color border_color;
