@@ -2,6 +2,7 @@
 #include <errno.h>
 #include <getopt.h>
 #include <locale.h>
+#include <poll.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -16,6 +17,7 @@
 #include <xkbcommon/xkbcommon.h>
 #include "tofi.h"
 #include "compgen.h"
+#include "drun.h"
 #include "config.h"
 #include "entry.h"
 #include "image.h"
@@ -30,6 +32,15 @@
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
 #define MIN(a, b) ((a) < (b) ? (a) : (b))
 
+static uint32_t gettime_ms() {
+	struct timespec t;
+	clock_gettime(CLOCK_MONOTONIC, &t);
+
+	uint32_t ms = t.tv_sec * 1000;
+	ms += t.tv_nsec / 1000000;
+	return ms;
+}
+
 static void zwlr_layer_surface_configure(
 		void *data,
 		struct zwlr_layer_surface_v1 *zwlr_layer_surface,
@@ -56,11 +67,6 @@ static void zwlr_layer_surface_configure(
 	tofi->window.surface.width = tofi->window.width;
 	tofi->window.surface.height = tofi->window.height;
 
-	/*
-	 * Need to redraw the background at the new size. This entails
-	 * a wl_surface_commit, so no need to do so explicitly here.
-	 */
-	tofi->window.surface.redraw = true;
 	zwlr_layer_surface_v1_ack_configure(
 			tofi->window.zwlr_layer_surface,
 			serial);
@@ -136,25 +142,12 @@ static void wl_keyboard_leave(
 	/* Deliberately left blank */
 }
 
-static void wl_keyboard_key(
-		void *data,
-		struct wl_keyboard *wl_keyboard,
-		uint32_t serial,
-		uint32_t time,
-		uint32_t key,
-		uint32_t state)
+static void handle_keypress(struct tofi *tofi, xkb_keycode_t keycode)
 {
-	struct tofi *tofi = data;
-	uint32_t keycode = key + 8;
 	if (tofi->xkb_state == NULL) {
 		return;
 	}
-	xkb_keysym_t sym = xkb_state_key_get_one_sym(
-			tofi->xkb_state,
-			keycode);
-	if (state != WL_KEYBOARD_KEY_STATE_PRESSED) {
-		return;
-	}
+	xkb_keysym_t sym = xkb_state_key_get_one_sym(tofi->xkb_state, keycode);
 
 	struct entry *entry = &tofi->window.entry;
 	char buf[5]; /* 4 UTF-8 bytes plus null terminator. */
@@ -165,7 +158,7 @@ static void wl_keyboard_key(
 			sizeof(buf));
 	wchar_t ch;
 	mbtowc(&ch, buf, sizeof(buf));
-	if (len > 0 && iswprint(ch) && !iswblank(ch)) {
+	if (len > 0 && iswprint(ch) && (entry->drun || !iswblank(ch))) {
 		if (entry->input_length < N_ELEM(entry->input) - 1) {
 			entry->input[entry->input_length] = ch;
 			entry->input_length++;
@@ -174,13 +167,36 @@ static void wl_keyboard_key(
 					buf,
 					N_ELEM(buf));
 			entry->input_mb_length += len;
-			struct string_vec tmp = entry->results;
-			entry->results = string_vec_filter(&entry->results, entry->input_mb);
-			string_vec_destroy(&tmp);
+			if (entry->drun) {
+				struct string_vec results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
+				string_vec_destroy(&entry->results);
+				entry->results = results;
+			} else {
+				struct string_vec tmp = entry->results;
+				entry->results = string_vec_filter(&entry->results, entry->input_mb, tofi->fuzzy_match);
+				string_vec_destroy(&tmp);
+			}
+		}
+	} else if (entry->input_length > 0
+			&& (sym == XKB_KEY_BackSpace
+				|| (sym == XKB_KEY_w
+					&& xkb_state_mod_name_is_active(
+						tofi->xkb_state,
+						XKB_MOD_NAME_CTRL,
+						XKB_STATE_MODS_EFFECTIVE))))
+	{
+		if (sym == XKB_KEY_BackSpace) {
+			entry->input_length--;
+			entry->input[entry->input_length] = L'\0';
+		} else {
+			while (entry->input_length > 0 && iswspace(entry->input[entry->input_length - 1])) {
+				entry->input_length--;
+			}
+			while (entry->input_length > 0 && !iswspace(entry->input[entry->input_length - 1])) {
+				entry->input_length--;
+			}
+			entry->input[entry->input_length] = L'\0';
 		}
-	} else if (entry->input_length > 0 && sym == XKB_KEY_BackSpace) {
-		entry->input_length--;
-		entry->input[entry->input_length] = L'\0';
 		const wchar_t *src = entry->input;
 		size_t siz = wcsrtombs(
 				entry->input_mb,
@@ -189,7 +205,28 @@ static void wl_keyboard_key(
 				NULL);
 		entry->input_mb_length = siz;
 		string_vec_destroy(&entry->results);
-		entry->results = string_vec_filter(&entry->commands, entry->input_mb);
+		if (entry->drun) {
+			entry->results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
+		} else {
+			entry->results = string_vec_filter(&entry->commands, entry->input_mb, tofi->fuzzy_match);
+		}
+	} else if (sym == XKB_KEY_u
+			&& xkb_state_mod_name_is_active(
+				tofi->xkb_state,
+				XKB_MOD_NAME_CTRL,
+				XKB_STATE_MODS_EFFECTIVE)
+		  )
+	{
+		entry->input_length = 0;
+		entry->input[0] = L'\0';
+		entry->input_mb_length = 0;
+		entry->input_mb[0] = '\0';
+		string_vec_destroy(&entry->results);
+		if (entry->drun) {
+			entry->results = desktop_vec_filter(&entry->apps, entry->input_mb, tofi->fuzzy_match);
+		} else {
+			entry->results = string_vec_filter(&entry->commands, entry->input_mb, tofi->fuzzy_match);
+		}
 	} else if (sym == XKB_KEY_Escape
 			|| (sym == XKB_KEY_c
 				&& xkb_state_mod_name_is_active(
@@ -200,28 +237,89 @@ static void wl_keyboard_key(
 		  )
 	{
 		tofi->closed = true;
+		return;
 	} else if (sym == XKB_KEY_Return || sym == XKB_KEY_KP_Enter) {
 		tofi->submit = true;
 		return;
 	}
 
-	if (sym == XKB_KEY_Up || sym == XKB_KEY_Left) {
-		uint32_t nsel = MAX(MIN(tofi->window.entry.num_results, tofi->window.entry.results.count), 1);
-		tofi->window.entry.selection += nsel;
-		tofi->window.entry.selection--;
-		tofi->window.entry.selection %= nsel;
-	} else if (sym == XKB_KEY_Down || sym == XKB_KEY_Right) {
-		uint32_t nsel = MAX(MIN(tofi->window.entry.num_results, tofi->window.entry.results.count), 1);
-		tofi->window.entry.selection++;
-		tofi->window.entry.selection %= nsel;
+	uint32_t nsel = MAX(MIN(entry->num_results_drawn, entry->results.count), 1);
+	if (sym == XKB_KEY_Up || sym == XKB_KEY_Left
+			|| (sym == XKB_KEY_k
+				&& xkb_state_mod_name_is_active(
+					tofi->xkb_state,
+					XKB_MOD_NAME_CTRL,
+					XKB_STATE_MODS_EFFECTIVE)
+			   )
+			) {
+		if (entry->selection > 0) {
+			entry->selection--;
+		} else {
+			if (entry->first_result > nsel) {
+				entry->first_result -= entry->last_num_results_drawn;
+				entry->selection = entry->last_num_results_drawn - 1;
+			} else if (entry->first_result > 0) {
+				entry->selection = entry->first_result - 1;
+				entry->first_result = 0;
+			}
+		}
+	} else if (sym == XKB_KEY_Down || sym == XKB_KEY_Right || sym == XKB_KEY_Tab
+			|| (sym == XKB_KEY_j
+				&& xkb_state_mod_name_is_active(
+					tofi->xkb_state,
+					XKB_MOD_NAME_CTRL,
+					XKB_STATE_MODS_EFFECTIVE)
+			   )
+			) {
+		entry->selection++;
+		if (entry->selection >= nsel) {
+			entry->selection -= nsel;
+			entry->first_result += nsel;
+			entry->first_result %= entry->results.count;
+			entry->last_num_results_drawn = entry->num_results_drawn;
+		}
 	} else {
-		tofi->window.entry.selection = 0;
+		entry->selection = 0;
+		entry->first_result = 0;
 	}
 	entry_update(&tofi->window.entry);
 	tofi->window.surface.redraw = true;
 	
 }
 
+static void wl_keyboard_key(
+		void *data,
+		struct wl_keyboard *wl_keyboard,
+		uint32_t serial,
+		uint32_t time,
+		uint32_t key,
+		uint32_t state)
+{
+	struct tofi *tofi = data;
+
+	/*
+	 * If this wasn't a keypress (i.e. was a key release), just update key
+	 * repeat info and return.
+	 */
+	uint32_t keycode = key + 8;
+	if (state != WL_KEYBOARD_KEY_STATE_PRESSED) {
+		if (keycode == tofi->repeat.keycode) {
+			tofi->repeat.active = false;
+		} else {
+			tofi->repeat.next = gettime_ms() + tofi->repeat.delay;
+		}
+		return;
+	}
+
+	/* A rate of 0 disables key repeat */
+	if (xkb_keymap_key_repeats(tofi->xkb_keymap, keycode) && tofi->repeat.rate != 0) {
+		tofi->repeat.active = true;
+		tofi->repeat.keycode = keycode;
+		tofi->repeat.next = gettime_ms() + tofi->repeat.delay;
+	}
+	handle_keypress(tofi, keycode);
+}
+
 static void wl_keyboard_modifiers(
 		void *data,
 		struct wl_keyboard *wl_keyboard,
@@ -251,7 +349,16 @@ static void wl_keyboard_repeat_info(
 		int32_t rate,
 		int32_t delay)
 {
-	/* Deliberately left blank */
+	struct tofi *tofi = data;
+	tofi->repeat.rate = rate;
+	tofi->repeat.delay = delay;
+	if (rate > 0) {
+		log_debug("Key repeat every %u ms after %u ms.\n",
+				1000 / rate,
+				delay);
+	} else {
+		log_debug("Key repeat disabled.\n");
+	}
 }
 
 static const struct wl_keyboard_listener wl_keyboard_listener = {
@@ -420,7 +527,13 @@ static void output_geometry(
 		const char *model,
 		int32_t transform)
 {
-	/* Deliberately left blank */
+	struct tofi *tofi = data;
+	struct output_list_element *el;
+	wl_list_for_each(el, &tofi->output_list, link) {
+		if (el->wl_output == wl_output) {
+			el->transform = transform;
+		}
+	}
 }
 
 static void output_mode(
@@ -489,8 +602,10 @@ static const struct wl_output_listener wl_output_listener = {
 	.mode = output_mode,
 	.done = output_done,
 	.scale = output_scale,
+#ifndef NO_WL_OUTPUT_NAME
 	.name = output_name,
 	.description = output_description,
+#endif
 };
 
 static void registry_global(
@@ -522,11 +637,18 @@ static void registry_global(
 		log_debug("Bound to seat %u.\n", name);
 	} else if (!strcmp(interface, wl_output_interface.name)) {
 		struct output_list_element *el = xmalloc(sizeof(*el));
+		if (version < 4) {
+			el->name = xstrdup("");
+			log_warning("Using an outdated compositor, "
+					"output selection will not work.\n");
+		} else {
+			version = 4;
+		}
 		el->wl_output = wl_registry_bind(
 				wl_registry,
 				name,
 				&wl_output_interface,
-				4);
+				version);
 		wl_output_add_listener(
 				el->wl_output,
 				&wl_output_listener,
@@ -541,11 +663,17 @@ static void registry_global(
 				1);
 		log_debug("Bound to shm %u.\n", name);
 	} else if (!strcmp(interface, zwlr_layer_shell_v1_interface.name)) {
+		if (version < 4) {
+			log_warning("Using an outdated compositor, "
+					"screen anchoring may not work.\n");
+		} else {
+			version = 4;
+		}
 		tofi->zwlr_layer_shell = wl_registry_bind(
 				wl_registry,
 				name,
 				&zwlr_layer_shell_v1_interface,
-				4);
+				version);
 		log_debug("Bound to zwlr_layer_shell_v1 %u.\n", name);
 	}
 }
@@ -646,7 +774,7 @@ static void usage()
 "Usage: tofi [options]\n"
 "\n"
 "  -h, --help                           Print this message and exit.\n"
-"  -c, --config                         Specify a config file.\n"
+"  -c, --config <path>                  Specify a config file.\n"
 "      --font <name|path>               Font to use.\n"
 "      --font-size <pt>                 Point size of text.\n"
 "      --background-color <color>       Color of the background.\n"
@@ -658,6 +786,10 @@ static void usage()
 "      --prompt-text <string>           Prompt text.\n"
 "      --num-results <n>                Maximum number of results to display.\n"
 "      --selection-color <color>        Color of selected result.\n"
+"      --selection-match-color <color>  Color of the matching portion of the\n"
+"                                       selected result.\n"
+"      --selection-padding <px>         Extra horizontal padding for selected\n"
+"                                       result background.\n"
 "      --selection-background <color>   Color of selected result background.\n"
 "      --result-spacing <px>            Spacing between results.\n"
 "      --min-input-width <px>           Minimum input width in horizontal mode.\n"
@@ -665,6 +797,7 @@ static void usage()
 "      --height <px|%>                  Height of the window.\n"
 "      --corner-radius <px>             Radius of window corners.\n"
 "      --output <name>                  Name of output to display window on.\n"
+"      --scale <true|false>             Follow the output's scale factor.\n"
 "      --anchor <position>              Location on screen to anchor window.\n"
 "      --margin-top <px|%>              Offset from top of screen.\n"
 "      --margin-bottom <px|%>           Offset from bottom of screen.\n"
@@ -677,6 +810,11 @@ static void usage()
 "      --hide-cursor <true|false>       Hide the cursor.\n"
 "      --horizontal <true|false>        List results horizontally.\n"
 "      --history <true|false>           Sort results by number of usages.\n"
+"      --fuzzy-match <true|false>       Use fuzzy matching for searching.\n"
+"      --drun-launch <true|false>       Launch apps directly in drun mode.\n"
+"      --drun-print-exec <true|false>   Print a command line in drun mode.\n"
+"                                       This is now always the case,\n"
+"                                       and this option is deprecated.\n"
 "      --hint-font <true|false>         Perform font hinting.\n"
 "      --late-keyboard-init             (EXPERIMENTAL) Delay keyboard\n"
 "                                       initialisation until after the first\n"
@@ -695,6 +833,8 @@ const struct option long_options[] = {
 	{"font-size", required_argument, NULL, 0},
 	{"num-results", required_argument, NULL, 0},
 	{"selection-color", required_argument, NULL, 0},
+	{"selection-match-color", required_argument, NULL, 0},
+	{"selection-padding", required_argument, NULL, 0},
 	{"selection-background", required_argument, NULL, 0},
 	{"outline-width", required_argument, NULL, 0},
 	{"outline-color", required_argument, NULL, 0},
@@ -717,34 +857,17 @@ const struct option long_options[] = {
 	{"horizontal", required_argument, NULL, 0},
 	{"hide-cursor", required_argument, NULL, 0},
 	{"history", required_argument, NULL, 0},
+	{"fuzzy-match", required_argument, NULL, 0},
+	{"drun-launch", required_argument, NULL, 0},
+	{"drun-print-exec", required_argument, NULL, 0},
 	{"hint-font", required_argument, NULL, 0},
-	{"output", required_argument, NULL, 'o'},
-	{"late-keyboard-init", no_argument, NULL, 'k'},
+	{"output", required_argument, NULL, 0},
+	{"scale", required_argument, NULL, 0},
+	{"late-keyboard-init", optional_argument, NULL, 'k'},
 	{NULL, 0, NULL, 0}
 };
 const char *short_options = ":hc:";
 
-static void parse_early_args(struct tofi *tofi, int argc, char *argv[])
-{
-	/* Handle errors ourselves (i.e. ignore them for now). */
-	opterr = 0;
-
-	/* Just check for help, late-keyboard-init and output */
-	optind = 1;
-	int opt = getopt_long(argc, argv, short_options, long_options, NULL);
-	while (opt != -1) {
-		if (opt == 'h') {
-			usage();
-			exit(EXIT_SUCCESS);
-		} else if (opt == 'k') {
-			tofi->late_keyboard_init = true;
-		} else if (opt == 'o') {
-			snprintf(tofi->target_output_name, N_ELEM(tofi->target_output_name), "%s", optarg);
-		}
-		opt = getopt_long(argc, argv, short_options, long_options, NULL);
-	}
-}
-
 static void parse_args(struct tofi *tofi, int argc, char *argv[])
 {
 
@@ -788,7 +911,17 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 	opt = getopt_long(argc, argv, short_options, long_options, &option_index);
 	while (opt != -1) {
 		if (opt == 0) {
-			apply_option(tofi, long_options[option_index].name, optarg);
+			config_apply(tofi, long_options[option_index].name, optarg);
+		} else if (opt == 'k') {
+			/*
+			 * Backwards compatibility for --late-keyboard-init not
+			 * taking an argument.
+			 */
+			if (optarg) {
+				config_apply(tofi, long_options[option_index].name, optarg);
+			} else {
+				tofi->late_keyboard_init = true;
+			}
 		}
 		opt = getopt_long(argc, argv, short_options, long_options, &option_index);
 	}
@@ -800,6 +933,41 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 	}
 }
 
+static void do_submit(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+	uint32_t selection = entry->selection + entry->first_result;
+	char *res = entry->results.buf[selection].string;
+	if (entry->drun) {
+		/*
+		 * TODO: This is ugly. The list of apps needs to be sorted
+		 * alphabetically for bsearch to find the selected entry, but
+		 * we previously sorted by history count. This needs fixing.
+		 */
+		desktop_vec_sort(&entry->apps);
+		struct desktop_entry *app = desktop_vec_find(&entry->apps, res);
+		if (app == NULL) {
+			log_error("Couldn't find application file! This shouldn't happen.\n");
+			return;
+		} else {
+			res = app->path;
+		}
+		if (tofi->drun_launch) {
+			drun_launch(res);
+		} else {
+			drun_print(res);
+		}
+	} else {
+		printf("%s\n", res);
+	}
+	if (tofi->use_history) {
+		history_add(
+				&entry->history,
+				entry->results.buf[selection].string);
+		history_save(&entry->history, entry->drun);
+	}
+}
+
 int main(int argc, char *argv[])
 {
 	/* Call log_debug to initialise the timers we use for perf checking. */
@@ -821,7 +989,6 @@ int main(int argc, char *argv[])
 				.font_name = "Sans",
 				.font_size = 24,
 				.prompt_text = "run: ",
-				.num_results = 5,
 				.padding_top = 8,
 				.padding_bottom = 8,
 				.padding_left = 8,
@@ -843,7 +1010,7 @@ int main(int argc, char *argv[])
 	};
 	wl_list_init(&tofi.output_list);
 
-	parse_early_args(&tofi, argc, argv);
+	parse_args(&tofi, argc, argv);
 
 	/*
 	 * Initial Wayland & XKB setup.
@@ -986,31 +1153,68 @@ int main(int argc, char *argv[])
 		 */
 		struct output_list_element *el;
 		el = wl_container_of(tofi.output_list.next, el, link);
-		tofi.output_width = el->width;
-		tofi.output_height = el->height;
+
+		/*
+		 * If we're rotated 90 degrees, we need to swap width and
+		 * height to calculate percentages.
+		 */
+		switch (el->transform) {
+			case WL_OUTPUT_TRANSFORM_90:
+			case WL_OUTPUT_TRANSFORM_270:
+			case WL_OUTPUT_TRANSFORM_FLIPPED_90:
+			case WL_OUTPUT_TRANSFORM_FLIPPED_270:
+				tofi.output_width = el->height;
+				tofi.output_height = el->width;
+				break;
+			default:
+				tofi.output_width = el->width;
+				tofi.output_height = el->height;
+		}
 		tofi.window.scale = el->scale;
+		tofi.window.transform = el->transform;
 		log_debug("Selected output %s.\n", el->name);
 	}
 
 	/*
-	 * We can now parse our arguments and config file, as we know the
-	 * output size required for specifying window sizes in percent.
+	 * We can now scale values and calculate any percentages, as we know
+	 * the output size and scale.
 	 */
-	parse_args(&tofi, argc, argv);
-
-	/* Scale fonts to the correct size. */
-	tofi.window.entry.font_size *= tofi.window.scale;
+	config_fixup_values(&tofi);
 
 	/*
 	 * If we were invoked as tofi-run, generate the command list.
+	 * If we were invoked as tofi-drun, generate the desktop app list.
 	 * Otherwise, just read standard input.
 	 */
 	if (strstr(argv[0], "-run")) {
 		log_debug("Generating command list.\n");
 		log_indent();
 		tofi.window.entry.commands = compgen_cached();
+		if (tofi.use_history) {
+			tofi.window.entry.history = history_load(tofi.window.entry.drun);
+			compgen_history_sort(&tofi.window.entry.commands, &tofi.window.entry.history);
+		}
 		log_unindent();
 		log_debug("Command list generated.\n");
+	} else if (strstr(argv[0], "-drun")) {
+		log_debug("Generating desktop app list.\n");
+		log_indent();
+		tofi.window.entry.drun = true;
+		struct desktop_vec apps = drun_generate_cached();
+		if (tofi.use_history) {
+			tofi.window.entry.history = history_load(tofi.window.entry.drun);
+			if (tofi.window.entry.drun) {
+				drun_history_sort(&apps, &tofi.window.entry.history);
+			}
+		}
+		struct string_vec commands = string_vec_create();
+		for (size_t i = 0; i < apps.count; i++) {
+			string_vec_add(&commands, apps.buf[i].name);
+		}
+		tofi.window.entry.commands = commands;
+		tofi.window.entry.apps = apps;
+		log_unindent();
+		log_debug("App list generated.\n");
 	} else {
 		char *line = NULL;
 		size_t n = 0;
@@ -1025,10 +1229,6 @@ int main(int argc, char *argv[])
 		free(line);
 		tofi.use_history = false;
 	}
-	if (tofi.use_history) {
-		tofi.window.entry.history = history_load();
-		compgen_history_sort(&tofi.window.entry.commands, &tofi.window.entry.history);
-	}
 	tofi.window.entry.results = string_vec_copy(&tofi.window.entry.commands);
 
 	/*
@@ -1062,7 +1262,7 @@ int main(int argc, char *argv[])
 			"launcher");
 	zwlr_layer_surface_v1_set_keyboard_interactivity(
 			tofi.window.zwlr_layer_surface,
-			ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND
+			ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE
 			);
 	zwlr_layer_surface_v1_add_listener(
 			tofi.window.zwlr_layer_surface,
@@ -1080,10 +1280,10 @@ int main(int argc, char *argv[])
 			tofi.window.height / tofi.window.scale);
 	zwlr_layer_surface_v1_set_margin(
 			tofi.window.zwlr_layer_surface,
-			tofi.window.margin_top,
-			tofi.window.margin_right,
-			tofi.window.margin_bottom,
-			tofi.window.margin_left);
+			tofi.window.margin_top / tofi.window.scale,
+			tofi.window.margin_right / tofi.window.scale,
+			tofi.window.margin_bottom / tofi.window.scale,
+			tofi.window.margin_left / tofi.window.scale);
 	wl_surface_commit(tofi.window.surface.wl_surface);
 
 	/*
@@ -1166,10 +1366,66 @@ int main(int argc, char *argv[])
 		tofi.late_keyboard_init = false;
 	}
 
-	while (wl_display_dispatch(tofi.wl_display) != -1) {
-		if (tofi.closed) {
-			break;
+	/*
+	 * Main event loop.
+	 * See the wl_display(3) man page for an explanation of the
+	 * order of the various functions called here.
+	 */
+	while (!tofi.closed) {
+		struct pollfd pollfd;
+		pollfd.fd = wl_display_get_fd(tofi.wl_display);
+
+		/* Make sure we're ready to receive events on the main queue. */
+		while (wl_display_prepare_read(tofi.wl_display) != 0) {
+			wl_display_dispatch_pending(tofi.wl_display);
 		}
+
+		/* Make sure all our requests have been sent to the server. */
+		while (wl_display_flush(tofi.wl_display) != 0) {
+			pollfd.events = POLLOUT;
+			poll(&pollfd, 1, -1);
+		}
+
+		/*
+		 * Set time to wait for poll() to -1 (unlimited), unless
+		 * there's some key repeating going on.
+		 */
+		int timeout = -1;
+		if (tofi.repeat.active) {
+			int64_t wait = (int64_t)tofi.repeat.next - (int64_t)gettime_ms();
+			if (wait >= 0) {
+				timeout = wait;
+			} else {
+				timeout = 0;
+			}
+		}
+
+		pollfd.events = POLLIN | POLLPRI;
+		int res = poll(&pollfd, 1, timeout);
+		if (res == 0) {
+			/*
+			 * No events to process and no error - we presumably
+			 * have a key repeat to handle.
+			 */
+			wl_display_cancel_read(tofi.wl_display);
+			if (tofi.repeat.active) {
+				int64_t wait = (int64_t)tofi.repeat.next - (int64_t)gettime_ms();
+				if (wait <= 0) {
+					handle_keypress(&tofi, tofi.repeat.keycode);
+					tofi.repeat.next += 1000 / tofi.repeat.rate;
+				}
+			}
+		} else if (res < 0) {
+			/* There was an error polling the display. */
+			wl_display_cancel_read(tofi.wl_display);
+		} else {
+			/* Events to read, so put them on the queue. */
+			wl_display_read_events(tofi.wl_display);
+		}
+
+		/* Handle any events we read. */
+		wl_display_dispatch_pending(tofi.wl_display);
+
 		if (tofi.window.surface.redraw) {
 			surface_draw(&tofi.window.surface);
 			tofi.window.surface.redraw = false;
@@ -1177,17 +1433,11 @@ int main(int argc, char *argv[])
 		if (tofi.submit) {
 			tofi.submit = false;
 			if (tofi.window.entry.results.count > 0) {
-				uint32_t selection = tofi.window.entry.selection;
-				printf("%s\n", tofi.window.entry.results.buf[selection]);
-				if (tofi.use_history) {
-					history_add(
-							&tofi.window.entry.history,
-							tofi.window.entry.results.buf[selection]);
-					history_save(&tofi.window.entry.history);
-				}
+				do_submit(&tofi);
 				break;
 			}
 		}
+
 	}
 
 	log_debug("Window closed, performing cleanup.\n");
@@ -1226,6 +1476,9 @@ int main(int argc, char *argv[])
 	xkb_keymap_unref(tofi.xkb_keymap);
 	xkb_context_unref(tofi.xkb_context);
 	wl_registry_destroy(tofi.wl_registry);
+	if (tofi.window.entry.drun) {
+		desktop_vec_destroy(&tofi.window.entry.apps);
+	}
 	string_vec_destroy(&tofi.window.entry.commands);
 	string_vec_destroy(&tofi.window.entry.results);
 	if (tofi.use_history) {
