@@ -3,6 +3,7 @@
 
 #include <stdbool.h>
 #include <stdint.h>
+#include <xkbcommon/xkbcommon.h>
 #include "color.h"
 #include "entry.h"
 #include "image.h"
@@ -18,6 +19,7 @@ struct output_list_element {
 	uint32_t width;
 	uint32_t height;
 	int32_t scale;
+	int32_t transform;
 };
 
 struct tofi {
@@ -53,17 +55,35 @@ struct tofi {
 		uint32_t width;
 		uint32_t height;
 		uint32_t scale;
+		int32_t transform;
 		int32_t margin_top;
 		int32_t margin_bottom;
 		int32_t margin_left;
 		int32_t margin_right;
+		bool width_is_percent;
+		bool height_is_percent;
+		bool margin_top_is_percent;
+		bool margin_bottom_is_percent;
+		bool margin_left_is_percent;
+		bool margin_right_is_percent;
 	} window;
+	struct {
+		uint32_t rate;
+		uint32_t delay;
+		uint32_t keycode;
+		uint32_t next;
+		bool active;
+	} repeat;
 
 	/* Options */
 	uint32_t anchor;
 	bool hide_cursor;
 	bool use_history;
+	bool use_scale;
 	bool late_keyboard_init;
+	bool drun_launch;
+	bool drun_print_exec;
+	bool fuzzy_match;
 	char target_output_name[MAX_OUTPUT_NAME_LEN];
 };
 
